from .refill_creator import RefillCreator


def create_refill(instance):
    """Creates the refill for this visit, if not already created.

    Called from signal.
    """
    number_of_days = 0
    if instance.subject_visit.appointment.next:
        number_of_days = (
            instance.subject_visit.appointment.next.appt_datetime
            - instance.subject_visit.appointment.appt_datetime
        ).days
    RefillCreator(
        subject_identifier=instance.subject_visit.subject_identifier,
        refill_date=instance.refill_date,
        visit_code=instance.subject_visit.appointment.visit_code,
        visit_code_sequence=instance.subject_visit.appointment.visit_code_sequence,
        number_of_days=number_of_days,
        dosage_guideline=instance.dosage_guideline,
        formulation=instance.formulation,
        make_active=True,
    )



# Generated by Django 3.2.11 on 2022-04-12 20:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('edc_pharmacy', '0002_alter_medication_unique_together'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='rxrefill',
            unique_together={('rx', 'refill_date'), ('rx', 'visit_code', 'visit_code_sequence')},
        ),
    ]



# Generated by Django 3.2.11 on 2022-04-15 14:47

from django.db import migrations
import edc_action_item.models.action_model_mixin
import edc_sites.models


class Migration(migrations.Migration):

    dependencies = [
        ('edc_pharmacy', '0004_auto_20220414_1741'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='rx',
            managers=[
                ('on_site', edc_sites.models.CurrentSiteManager()),
                ('objects', edc_action_item.models.action_model_mixin.ActionItemModelManager()),
            ],
        ),
    ]



from dateutil.relativedelta import relativedelta
from django.test import TestCase, tag
from edc_constants.constants import FEMALE
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow

from edc_pharmacy.exceptions import PrescriptionNotStarted
from edc_pharmacy.models import (
    DosageGuideline,
    Formulation,
    FormulationType,
    FrequencyUnits,
    Medication,
    Route,
    Rx,
    RxRefill,
    Units,
)
from edc_pharmacy.refill import RefillCreator, get_active_refill
from edc_pharmacy.refill.refill_creator import RefillAlreadyExists


@tag("refill")
class TestRefill(TestCase):
    def setUp(self):
        self.subject_identifier = "12345"

        RegisteredSubject.objects.create(
            subject_identifier="12345",
            dob=get_utcnow() - relativedelta(years=25),
            gender=FEMALE,
        )

        self.medication = Medication.objects.create(
            name="Flucytosine",
            display_name="flucytosine",
        )

        # create prescription for this subject
        self.rx = Rx.objects.create(
            subject_identifier=self.subject_identifier,
            weight_in_kgs=40,
            report_datetime=get_utcnow(),
        )
        self.rx.medications.add(self.medication)

        self.formulation = Formulation.objects.create(
            medication=self.medication,
            strength=500,
            units=Units.objects.get(name="mg"),
            route=Route.objects.get(display_name="Oral"),
            formulation_type=FormulationType.objects.get(display_name__iexact="Tablet"),
        )

        self.dosage_guideline = DosageGuideline.objects.create(
            medication=self.medication,
            dose_per_kg=100,
            dose_units=Units.objects.get(name="mg"),
            frequency=1,
            frequency_units=FrequencyUnits.objects.get(name="day"),
        )

    def test_rx_refill_str(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            visit_code="1000",
            visit_code_sequence=0,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            frequency=1,
            refill_date=get_utcnow(),
            number_of_days=10,
        )
        self.assertTrue(str(obj))

    def test_prescription_accepts_explicit_dose(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            visit_code="1000",
            visit_code_sequence=0,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            frequency=1,
            dose=3,
            refill_date=get_utcnow(),
            number_of_days=10,
        )
        self.assertEqual(obj.dose, 3)

    def test_prescription_calculates_dose(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            visit_code="1000",
            visit_code_sequence=0,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            frequency=1,
            dose=None,
            refill_date=get_utcnow(),
            number_of_days=10,
        )
        self.assertEqual(obj.dose, 8.0)
        self.assertEqual(obj.medication.units, "mg")

    def test_prescription_total(self):
        obj = RxRefill.objects.create(
            rx=self.rx,
            visit_code="1000",
            visit_code_sequence=0,
            formulation=self.formulation,
            dosage_guideline=self.dosage_guideline,
            frequency=1,
            dose=None,
            refill_date=get_utcnow(),
            number_of_days=10,
        )
        self.assertEqual(obj.total, 56)

    def test_refill_gets_rx(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )
        self.assertTrue(refill_creator.refill.rx)

    def test_refill_raises_on_gets_rx(self):
        """Assert raises if refill date before Rx"""
        self.assertRaises(
            PrescriptionNotStarted,
            RefillCreator,
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=(get_utcnow() - relativedelta(years=1)).date(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )

    def test_refill_create_and_no_active_refill(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=False,
        )
        self.assertIsNone(get_active_refill(refill_creator.refill.rx))

    def test_refill_create_and_gets_active_refill(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )
        self.assertEqual(
            get_active_refill(refill_creator.refill.rx),
            refill_creator.refill,
        )

    def test_refill_create_activates_by_default(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow().date(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )
        self.assertTrue(get_active_refill(refill_creator.refill.rx).active)

    def test_refill_create_does_not_activate_if_false(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow().date(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=False,
        )
        self.assertIsNone(get_active_refill(refill_creator.refill.rx))

    def test_refill_create_duplicate_raises(self):
        RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )

        self.assertRaises(
            RefillAlreadyExists,
            RefillCreator,
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )

    def test_refill_create_finds_active(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=True,
        )
        self.assertIsNotNone(get_active_refill(refill_creator.refill.rx))
        refill_creator.refill.deactivate()
        self.assertIsNone(get_active_refill(refill_creator.refill.rx))

    def test_refill_create_activates_next(self):
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=get_utcnow(),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )
        self.assertIsNotNone(get_active_refill(refill_creator.refill.rx))

        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1010",
            visit_code_sequence=0,
            refill_date=get_utcnow() + relativedelta(months=1),
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=True,
        )
        self.assertIsNotNone(get_active_refill(refill_creator.refill.rx))
        self.assertEqual(get_active_refill(refill_creator.refill.rx), refill_creator.refill)

    def test_refill_create_refill_date(self):
        refill_date = get_utcnow().date()
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=refill_date,
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
        )

        self.assertEqual(refill_creator.refill.refill_date, refill_date)

    def test_refill_create_and_make_active(self):
        refill_date = get_utcnow().date()
        refill_creator = RefillCreator(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
            visit_code_sequence=0,
            refill_date=refill_date,
            number_of_days=32,
            dosage_guideline=self.dosage_guideline,
            formulation=self.formulation,
            make_active=False,
        )
        self.assertFalse(refill_creator.refill.active)
        refill_creator.refill.activate()
        self.assertTrue(refill_creator.refill.active)
        self.assertEqual(RxRefill.objects.all().count(), 1)



from typing import Any

from django.core.exceptions import ObjectDoesNotExist

from ..models import Rx, RxRefill


def get_active_refill(rx: Rx) -> Any:
    """Returns the 'active' Refill instance or None
    for this prescription.

    This does not return a model instance.
    """
    try:
        obj = RxRefill.objects.get(rx=rx, active=True)
    except ObjectDoesNotExist:
        return None
    return Refill(obj)


class Refill:
    """A wrapper class of the RxRefill model."""

    def __init__(self, rx_refill: RxRefill):
        self._object = rx_refill

    def __repr__(self):
        return f"{self.__class__.__name__}({self._object})"

    def __str__(self):
        return f"{self.__class__.__name__}({self._object})"

    def __eq__(self, other):
        if isinstance(other, self.__class__) and self._object == other._object:
            return True
        return False

    @property
    def rx(self):
        return self._object.rx

    @property
    def refill_date(self):
        return self._object.refill_date

    @property
    def active(self):
        return self._object.active

    @property
    def dispensed(self):
        return self._object.dispensed

    @property
    def remaining(self):
        return self._object.remaining

    def activate(self):
        """Activates this refill and deactivates the active refill
        if there is one.
        """
        if active_refill := get_active_refill(self._object.rx):
            if active_refill != self:
                active_refill.deactivate()
        self._object.active = True
        self._object.save()
        self._object.refresh_from_db()

    def deactivate(self):
        """Deactivates this refill."""
        if self._object.active:
            self._object.active = False
            self._object.save()
            self._object.refresh_from_db()



from django.core.exceptions import ObjectDoesNotExist
from edc_constants.constants import NO, YES
from edc_form_validators import INVALID_ERROR, FormValidator

from ..models import Rx, RxRefill


class StudyMedicationFormValidator(FormValidator):
    def clean(self):
        self.required_if(YES, field="order_next", field_required="next_dosage_guideline")
        if self.cleaned_data.get("order_next") == NO and self.next_refill:
            if self.next_refill.active:
                self.raise_validation_error(
                    "Invalid. Next refill is already active", INVALID_ERROR
                )
        if (
            self.cleaned_data.get("order_next") == NO
            and not self.cleaned_data.get("subject_visit").appointment.next
        ):
            self.raise_validation_error(
                "Invalid. This is the last scheduled visit", INVALID_ERROR
            )

        self.required_if(YES, field="order_next", field_required="next_formulation")

    @property
    def next_refill(self):
        for obj in RxRefill.objects.filter(
            rx=self.rx,
            refill_date__gt=self.cleaned_data.get("refill_date"),
        ).order_by("refill_date"):
            return obj
        return None

    @property
    def rx(self):
        try:
            return Rx.objects.get(
                subject_identifier=self.cleaned_data.get("subject_visit").subject_identifier,
                medications__in=[self.cleaned_data.get("formulation").medication],
            )
        except ObjectDoesNotExist:
            self.raise_validation_error(
                {"__all__": "Prescription does not exist"}, INVALID_ERROR
            )



from .dispensing_history import DispensingHistory
from .dosage_guideline import DosageGuideline
from .formulation import Formulation
from .list_models import Container, FormulationType, FrequencyUnits, Route, Units
from .medication import Medication
from .medication_lot import MedicationLot
from .medication_order import MedicationOrder
from .medication_product import MedicationProduct
from .medication_stock import MedicationStock
from .medication_stock_create_labels import Labels, MedicationStockCreateLabels
from .medication_stock_receiving import MedicationStockReceiving
from .model_mixins import (
    MedicationOrderModelMixin,
    StudyMedicationCrfModelMixin,
    StudyMedicationRefillModelMixin,
)
from .proxy_models import VisitSchedule
from .return_history import ReturnError, ReturnHistory
from .rx import Rx
from .rx_refill import RxRefill
from .signals import dispensing_history_on_post_save
from .subject import Subject



from decimal import Decimal
from typing import Any, Optional, Union


class DosageError(Exception):
    pass


def dosage_per_day(
    dosage_guideline: Any,
    strength: Optional[Union[float, Decimal]] = None,
    strength_units: Optional[str] = None,
    weight_in_kgs: Optional[Union[float, Decimal]] = None,
):
    strength = strength or 1.0
    weight_in_kgs = weight_in_kgs or 1.0
    if strength_units and strength_units != dosage_guideline.dose_units.name:
        raise DosageError(
            f"Invalid units. Guideline dose is in "
            f"'{dosage_guideline.dose_units}'. Got {strength_units}."
        )
    return (
        float(dosage_guideline.dose or dosage_guideline.dose_per_kg)
        * float(weight_in_kgs)
        * float(dosage_guideline.frequency)
    ) / float(strength)
